/*
	LICENSE		Placed in public domain by cbonar@users.sf.net, 2002.  Share and enjoy!
	AUTHOR		cbonar@users.sf.net
	PROJECT		http://subtools.sourceforge.net
	CVS		cvs.subtools.sourceforge.net

	TODO :	use tmpfile() instead of tmpnam() -> implies to find another way to copy to the destination file (don't use cp() anymore)
		BUT test it because there was a bug creating the tmpfile on a cross-linked device...
*/

/* A list of the different #define used in this program */
/*#define NDEBUG*/		/* no debug nor assertion if defined */
/*#define SRTSHIFT*/		/* compile this program with the only ability to shift subtitles */
/*#define SRTRATE*/		/* compile this program with the only ability to change the rate of the subtitles */


#include <stdio.h>		/* FILE ... */
#ifdef sun
#include <unistd.h>		/* unlink for Solaris */
#endif
#ifdef unix
#include <unistd.h>		/* for some unices */
#endif
#include <string.h>		/* strlen, strcmp, strerror, ... */
#include <errno.h>		/* errno, error constants */


#define VERSION "20051215alpha"	/* Version of this program */
#define SZ_LINE 4096		/* Maximum length of a single line of input file, in characters */

/* easier with that struct */
typedef struct
{
        int hour;
        int min;
        int sec;
        int msec;
} tTimeStamp;




/*
	Duplicates a file.
*/
int cp( char *src, char *dest )
{
	FILE *fsrc,*fdest;
	char buffer[1024];		/* sizeable */

	fsrc = fopen(src,"r");
	if ( fsrc == NULL )
	{
		fprintf(stderr,"Error while opening [%s] : [%s]\n",src,strerror(errno));
		return -2;
	}
	fdest = fopen(dest,"w");
	if ( fdest == NULL )
	{
		fprintf(stderr,"Error while opening [%s] : [%s]\n",dest,strerror(errno));
		fclose(fsrc);
		return -3;
	}

	while ( fread(buffer,1,1,fsrc)==1 )
	{
		if ( fwrite(buffer,1,1,fdest)!=1 )
		{
			fprintf(stderr,"Error while writing to destination file [%s] -> [%s] : [%s]\n",src,dest,strerror(errno));
			return -1;
		}
	}

	fclose(fdest);
	fclose(fsrc);

	return 0;
}



/*
	OpTimeStamp : increase or decrease the given time stamp

	a     = initial time stamp ; is modified
	op    = type of operation to apply to a
	delay = amount of time (parameter of op)
	rate  = multiplier

	-> -1 : unknow operator
	-> 0  : ok
*/
#ifdef SRTSHIFT
int OpTimeStamp( tTimeStamp *a, char op, unsigned long delay )
#elif SRTRATE
int OpTimeStamp( tTimeStamp *a, float rate )
#endif
{
	unsigned long mtime;

	/* mtime is the amount of time represented by a, in milliseconds */
	mtime = a->hour*60*60*1000 + a->min*60*1000 + a->sec*1000 + a->msec;

	#ifdef SRTSHIFT
	switch ( op )
	{
		/* Increases the time stamp */
		case '+':
			/* data lost */
			if ( mtime + delay > 999 + 59*1000 + 59*60*1000 + 99*60*60*1000 ) /* max time stamp */
			{
				a->msec = 999;
				a->sec = 59;
				a->min = 59;
				a->hour = 99;
			}
			else
			{
				mtime += delay;
				a->msec = mtime % 1000;
				mtime /= 1000;
				a->sec = mtime % 60;
				mtime /= 60;
				a->min = mtime % 60;
				mtime /= 60;
				a->hour = mtime;        /* it is an amount of hours, not an hour */
			}
			break;

		/* Decreases the time stamp */
		case '-':
			if ( mtime < delay ) /* min time stamp */
			{
				a->msec = 000;
				a->sec = 00;
				a->min = 00;
				a->hour = 00;
			}
			else
			{
				/* same as for '+' */
				mtime -= delay;
				a->msec = mtime % 1000;
				mtime /= 1000;
				a->sec = mtime % 60;
				mtime /= 60;
				a->min = mtime % 60;
				mtime /= 60;
				a->hour = mtime;
			}
			break;

		default:
			fprintf(stderr,"OpTimeStamp: Unknown operator : [%c]\n",op);
			return -1;
	}
	#elif SRTRATE
		mtime *= rate;		/* what a cast... */
		a->msec = mtime % 1000;
		mtime /= 1000;
		a->sec = mtime % 60;
		mtime /= 60;
		a->min = mtime % 60;
		mtime /= 60;
		a->hour = mtime;
	#endif

	return 0;
}




/*
	AdjustTimeStamp : apply the modification to a time stamp

	line  = null-T string containing the time stamp (read from the command line)
			must be something like : "00:01:14,057 --> 00:01:16,571"
	op    = type of operation to apply to the time stamp
	delay = amount of time (parameter of the operation)
	rate  = the multiplier

	-> -1 : OpTimeStamp error on the first part
	-> -2 : OpTimeStamp error on the second part
	-> 0  : ok
*/
#ifdef SRTSHIFT
int AdjustTimeStamp( char *line, char op, unsigned long delay )
#elif SRTRATE
int AdjustTimeStamp( char *line, float rate )
#endif
{
	tTimeStamp from,to;

	/* read time stamps from the line */
	sscanf(line,"%u:%u:%u,%u --> %u:%u:%u,%u",
		&from.hour,&from.min,&from.sec,&from.msec,
		&to.hour,&to.min,&to.sec,&to.msec);

	/* changes the values of the time stamps */
	#ifdef SRTSHIFT
	if ( OpTimeStamp(&from,op,delay) )
		return -1;
	if ( OpTimeStamp(&to,op,delay) )
		return -2;
	#elif SRTRATE
	if ( OpTimeStamp(&to,rate) )
		return -1;
	if ( OpTimeStamp(&from,rate) )
		return -2;
	#endif

	/* write new time stamps to the line */
	sprintf(line,"%2.2u:%2.2u:%2.2u,%3.3u --> %2.2u:%2.2u:%2.2u,%3.3u\n",
		from.hour,from.min,from.sec,from.msec,
		to.hour,to.min,to.sec,to.msec);

	return 0;
}



/*
	GetDelay : converts a Subripper time stamp to an amount of time

	op   = operator to be written
	delay = amount of time (parameter of op)
	arg = null-T string containing the time stamp (read from the command line)

	-> 0L & op='?' : line is too short
	-> 0L & op='!' : line doesnt contain a valid time stamp
	-> 0L          : time stamp set to zero
	-> >0L         : amount of time represented by the time stamp
*/
unsigned long GetDelay( char *op, unsigned long *delay, char *arg )
{
	tTimeStamp fmr;

	/* avoid core dump */
	if ( strlen(arg) != 13 )
	{
		*op = '?';
		return (*delay = 0L);
	}

	/* most checking is done here by sscanf */
	if ( sscanf(arg,"%c%u:%u:%u,%u",op,&fmr.hour,&fmr.min,&fmr.sec,&fmr.msec) != 5 )
	{
		*op = '!';
		return (*delay = 0L);
	}

	#ifndef NDEBUG
	fprintf(stderr,"GetDelay: arg=[%s]\n",arg);
	fprintf(stderr,"GetDelay: op=[%c]\n",*op);
	fprintf(stderr,"GetDelay: timestamp=[%u][%u][%u][%u]\n",fmr.hour,fmr.min,fmr.sec,fmr.msec);
	#endif

	return (*delay = fmr.hour*60*60*1000 + fmr.min*60*1000 + fmr.sec*1000 + fmr.msec);
}



/*
	We don't accept null nor negative values.

	-> -1   : error on the first rate
	-> -2   : error on the second rate
	-> >=0    : ok, rate filled
*/
int GetRate( float *rate, char *arg1, char *arg2 )
{
	float fromRate;
	float toRate;

	/* We should not need to check for rate<0, but we need for rate!=0 anyway (true ?) */
	if ( sscanf(arg1,"%f",&fromRate) != 1 || fromRate <= 0 )
		return -1;
	if ( sscanf(arg2,"%f",&toRate) != 1 || toRate <= 0 )
		return -2;

	*rate = fromRate / toRate;	/* cannot throw an error because toRate!=0 is checked before */

	#ifndef NDEBUG
	fprintf(stderr,"GetRate: args=[%s] [%s]\n",arg1,arg2);
	fprintf(stderr,"GetRate: fromRate=[%f]\n",fromRate);
	fprintf(stderr,"GetRate: toRate=[%f]\n",toRate);
	fprintf(stderr,"GetRate: rate=[%f]\n",*rate);
	#endif

	return 0;
}




/*
	Init : initializes parameters from command-line arguments

	fin,fout  = input/output streams ; modified
	delay     = amount of time (parameter of op) ; modified
	delayOp   = type of operation to apply to the time stamp ; modified
	rate      = multiplier to apply to the time stamps ; modified
	ftmp   = name of a temporary swap file ; modified
	argc,argv = main() arguments

	-> -1 : cannot open input file
	-> -2 : cannot create temp file
	-> -3 : cannot create output file
	-> -4 : unable to get the time stamp modifier
	-> 0  : ok, all parameters done
*/
#ifdef SRTSHIFT
int Init( FILE **fin, FILE **fout, char **in, char **out, unsigned long *delay, char *delayOp, char **ftmp, int argc, char **argv )
#elif SRTRATE
int Init( FILE **fin, FILE **fout, char **in, char **out, float *rate, char **ftmp, int argc, char **argv )
#endif
{
	int arg = 1;	/* cursor on the args */

	#ifndef NDEBUG
	int i;
	for ( i=0 ; i<argc ; i++ )
		fprintf(stderr,"Init: argv[%d]=[%s]\n",i,argv[i]);
	#endif

	/* choose temporary file */
	/* TODO : and if we don't need one ? */
	*ftmp = tempnam(NULL,"subto");	/* second argument is only 5 char long */
	if ( *ftmp == NULL )
	{
		fprintf(stderr,"Error while creating temporary file : %s.\n",strerror(errno));
		return -2;
	}

	#ifdef SRTSHIFT
	/* init delay */
	if ( GetDelay(delayOp,delay,argv[arg])==0L )
	{
		if ( *delayOp=='?' || *delayOp=='!' )
		{
			fprintf(stderr,"Error in the time stamp pattern : [%s]; should be like [XHH:MM:SS,mmm]\n",argv[arg]);
			return -4;
		}
		/* else, do a simple file copy */
	}
	arg += 1;
	#elif SRTRATE
	if ( GetRate(rate,argv[arg],argv[arg+1])<0 )
	{
		fprintf(stderr,"Error in the rates : [%s] and [%s] should be float values\n",argv[arg],argv[arg+1]);
		return -4;
	}
	arg += 2;
	#endif

	/* init input */
	if ( !strcmp(argv[arg],"-") )
		*fin = stdin;
	else
	{
		*fin = fopen(argv[arg],"r");
		if ( *fin==NULL )
		{
			fprintf(stderr,"Error while trying to open input file [%s] : %s.\n",argv[arg],strerror(errno));
			return -1;
		}
		else
			*in = argv[arg];
	}
	arg += 1;

	/* init output */
	if ( arg >= argc )
		*fout = stdout;
	else
	{
		*out = argv[arg];
		/* don't open directly the destination file, because it is allowed to be be the same */
		*fout = fopen(*ftmp,"w");     /* not using ftmp() to rename the temporary file at the end */
		if ( *fout==NULL )
		{
			fprintf(stderr,"Error while trying to create output file [%s] : %s.\n",*ftmp,strerror(errno));
			return -3;
		}
	}

	return 0;
}




/*
	Clean : cleans initialized variables and return code

	Call it right before to exit.
	If an error happens in this function, it returns 127 less code.
*/
int Clean( int code, FILE *fin, FILE *fout, char *in, char *out, char *ftmp, int argc, char **argv )
{
	/* not necessary, but I feel better ;) */
	if ( fin != NULL )
		fclose(fin);
	if ( fout != NULL )
	{
		fclose(fout);

		/* lastly, delete the temporary file */
		if ( out != NULL )
		{
			/* delete destination file if exists */
			if ( unlink(out) )
			{
				switch ( errno )
				{
					case ENOENT:
						break;
					default:
						fprintf(stderr,"Error while deleting file [%s] : %s.\n",out,strerror(errno));
						return 127-code;
				}
			}

			/* copy the temp file into the destination directory */
			if ( cp(ftmp,out) )		/* needed to pass through cross-linked devices */
			{
				fprintf(stderr,"Your file has been output to [%s]\n",ftmp);
				return 127-code;
			}
			else
			{
				/* delete temp file only if it has been successfully duplicated */
				if ( unlink(ftmp) )
					/* don't exit, but warn the user */
					fprintf(stderr,"Warning : Error while deleting file [%s] : %s.\n",ftmp,strerror(errno));
			}
		}
	}

	return code;
}





/*
	Prints the usage page.

	out  = the stream where to print into
*/
void usage( FILE *out, int argc, char **argv )
{
	#ifdef SRTSHIFT
	fprintf(out,"Shifts the time stamps of a SubRipper subtitles file (.srt) to the beginning or the end.\n",argv[0]);
	#elif SRTRATE
	fprintf(out,"Adjusts the frame rate of the time stamps of a SubRipper subtitles file (.srt).\n",argv[0]);
	#endif
	
	fprintf(out,"\nUSAGE\n");
	#ifdef SRTSHIFT
	fprintf(out,"  %s delay_pattern source_file [dest_file]\n",argv[0]);
	#elif SRTRATE
	fprintf(out,"  %s source_rate dest_rate source_file [dest_file]\n",argv[0]);
	#endif
	
	fprintf(out,"\nOPTIONS\n");
	#ifdef SRTSHIFT
	fprintf(out,"  delay_pattern xHH:MM:SS,mmm\n");
	fprintf(out,"                x = + | - ; + shifts the subtitles to the end\n");
	fprintf(out,"                HH:MM:SS,mm = hour:minutes:seconds,milliseconds\n");
	#elif SRTRATE
	fprintf(out,"  source_rate   float value (dotted)\n");
	fprintf(out,"  dest_rate     float value (dotted)\n");
	#endif
	fprintf(out,"  source_file   SRT file to adjust\n");
	fprintf(out,"                if source_file is '-', data is read from the standard input\n");
	fprintf(out,"  dest_file     modified SRT file\n");
	fprintf(out,"                if not specified, the standard output is used\n");
	
	fprintf(out,"\nEXAMPLE\n");
	#ifdef SRTSHIFT
	fprintf(out,"  %s +00:00:01,500 matrix.srt matrix_better.srt\n",argv[0]);
	#elif SRTRATE
	fprintf(out,"  %s 23.976 25.00 matrix.srt matrix_better.srt\n",argv[0]);
	#endif
	
	fprintf(out,"\nRELEASE\n");
	fprintf(out,"  %s by cbonar@users.sf.net\n",VERSION);
	fprintf(out,"  Visit http://subtools.sourceforge.net\n");
}




/*
	main : Adjusts the time stamps within a Subripper file (.srt)

	-> 1 : missing parameter
	-> 2 : erroneous parameter (Init() error)
	-> 3 : error writting to output (fputs() error)
	-> 4 : AdjustTimeStamp() error (wrong operator, should never be)
	-> 0 : ok, done
*/
int main( int argc, char **argv )
{
	FILE *fin=NULL,*fout=NULL;	/* input and output streams */
	char *in=NULL, *out=NULL;	/* input and output files names */
	unsigned long current,next;	/* markers for counting the number of subtitles set */
	char buffin[SZ_LINE];		/* read buffer for fgets() */
	#ifdef SRTSHIFT
	unsigned long delay;		/* amount of time to change time stamps */
	char delayOp;			/* adding or substracting delay to time stamps */
	#elif SRTRATE
	float rate;			/* multiplier to apply to the time stamps, to get the correct rate */
	#endif
	char *ftmp=NULL;		/* temporary file for output */


	/* usage if necessary */
	#ifdef SRTSHIFT
	if ( argc < 3 )
	#elif SRTRATE
	if ( argc < 4 )
	#endif
	{
		usage(stderr,argc,argv);
		return Clean(1,fin,fout,in,out,ftmp,argc,argv);
	}
	else
	{
		/* first, inits */
		#ifdef SRTSHIFT
		if ( Init(&fin,&fout,&in,&out,&delay,&delayOp,&ftmp,argc,argv) )
		#elif SRTRATE
		if ( Init(&fin,&fout,&in,&out,&rate,&ftmp,argc,argv) )
		#endif
			return Clean(2,fin,fout,in,out,ftmp,argc,argv);

		/* then, does the stuff */
		current = 0L;
		/* search for the first subtitle set */
		do
		{
			if ( fgets(buffin,SZ_LINE,fin) == NULL )
				break;
			sscanf(buffin,"%u",&next);
		} while ( next != current+1L );
		current++;

		do
		{
			/* last set number read */
			if ( fputs(buffin,fout) < 0 )
				return Clean(3,fin,fout,in,out,ftmp,argc,argv);

			/* time stamp */
			if ( fgets(buffin,SZ_LINE,fin) == NULL )
				break;
			#ifdef SRTSHIFT
			if ( AdjustTimeStamp(buffin,delayOp,delay) )
			#elif SRTRATE
			if ( AdjustTimeStamp(buffin,rate) )
			#endif
				return Clean(4,fin,fout,in,out,ftmp,argc,argv);
			if ( fputs(buffin,fout) < 0 )
				return Clean(3,fin,fout,in,out,ftmp,argc,argv);

			/* subtitles and next line number */
			while ( 1 )
			{
				if ( fgets(buffin,SZ_LINE,fin) == NULL )
					break;
				sscanf(buffin,"%u",&next);
				if ( next == current+1L )
				{
					/* this seems to be the next set ; let's loop */
					current++;
					break;
				}
				else
				{
					if ( fputs(buffin,fout) < 0 )
						return Clean(3,fin,fout,in,out,ftmp,argc,argv);
				}
			}
		} while ( 1 );


	/* finally, exits */
	return Clean(0,fin,fout,in,out,ftmp,argc,argv);
	}
}
